local MINOR_VERSION = tonumber(("$Revision: 72 $"):match("%d+"))
local L = LibStub("AceLocale-3.0"):GetLocale("QuestAgent", false)
QuestAgent = LibStub("AceAddon-3.0"):NewAddon("QuestAgent", "AceConsole-3.0", "AceComm-3.0", "AceSerializer-3.0", "AceEvent-3.0", "AceTimer-3.0", "AceHook-3.0")
local QuestAgent = QuestAgent
QuestAgent.version = MINOR_VERSION
local tablet = AceLibrary("Tablet-2.0")
local dewdrop = AceLibrary("Dewdrop-2.0")
local compress = LibStub("LibCompress")
local ldb = LibStub("LibDataBroker-1.1")
local lbz = LibStub("LibBabble-Zone-3.0"):GetReverseLookupTable()
--local dbicon = LibStub("LibDBIcon-1.0")
---


--[[-- LOCALS ------------------------------------------------------------------------------------------------------------------]]

local WHT = "|cffffffff"
local BLK = "|cff000000"
local YEL = "|cffffff20"
local RED = "|cffff2020"
local GRN = "|cff20ff20"
local BLU = "|cff2020ff"
local HIL = "|cfffbd284"

local log = {}
local exp = {}
local pet = {}
local ver = {}
local ach = {}
local menu
local throttleTime = 60 -- Minimum delay between full updates, in seconds
local throttleTimeDiff = 3 -- Minimum delay between differential updates, in seconds
local last = 0
local lastDiff = 0
local inThrottle = false
local inThrottleDiff = false
local debug = false
local myquests
local versionWarned = false
local brokerPlugin
local tabletRegistered
local tipFrame
local commPrefix = "QAGNT" -- Also update this in QuestFrame.lua if changed!
local inTipRefresh = false



--[[-- MISTS OF PANDARIA RAID/PARTY API COMPATIBILITY LAYER --------------------------------------------------------------------]]

local IsRaidLeader = _G.IsRaidLeader
local IsPartyLeader = _G.IsPartyLeader
local UnitIsPartyLeader = _G.UnitIsPartyLeader
if UnitIsGroupLeader ~= nil then
 IsRaidLeader = function(unit) if UnitIsGroupLeader(unit) then return 1 else return nil end end
 IsPartyLeader = function(unit) if UnitIsGroupLeader(unit) then return 1 else return nil end end
 UnitIsPartyLeader = function(unit) if UnitIsGroupLeader(unit) then return 1 else return nil end end
end
local GetNumPartyMembers = _G.GetNumPartyMembers
local GetRealNumPartyMembers = _G.GetRealNumPartyMembers
if GetNumSubgroupMembers ~= nil then
 GetNumPartyMembers = function() return GetNumSubgroupMembers() end
 GetRealNumPartyMembers = function() return GetNumSubgroupMembers(LE_PARTY_CATEGORY_HOME) end
end
local GetNumRaidMembers = _G.GetNumRaidMembers
local GetRealNumRaidMembers = _G.GetRealNumRaidMembers
if GetNumGroupMembers ~= nil then
 GetNumRaidMembers = function() return GetNumGroupMembers() end
 GetRealNumRaidMembers = function() return GetNumGroupMembers(LE_PARTY_CATEGORY_HOME) end
end


--[[-- INITIALIZATION ----------------------------------------------------------------------------------------------------------]]

function QuestAgent:OnInitialize()

 -- Initialize options
	menu = {
		name = "QuestAgent",
		type = "group",
		args = {
		 header1 = {
			 type = "header",
			 name = L["Options"],
			 order = 10,
		 },
			reportLogChanges = {
				type = "toggle",
				name = L["Report new party quests and turn-ins"],
				desc = L["reportLogChanges"],
				get = function() return QuestAgent.db.profile.reportLogChanges end,
				set = function() QuestAgent.db.profile.reportLogChanges = not QuestAgent.db.profile.reportLogChanges end,
			 order = 11,
			},
			showPetExp = {
				type = "toggle",
				name = L["Show pet experience"],
				desc = L["showPetExp"],
				get = function() return QuestAgent.db.profile.showPetExp end,
				set = function()
				 QuestAgent.db.profile.showPetExp = not QuestAgent.db.profile.showPetExp
				 QuestAgent:UpdateDisplay()
				end,
			 order = 11,
			},
			reportDings = {
				type = "toggle",
				name = L["Report party level changes"],
				desc = L["reportDings"],
				get = function() return QuestAgent.db.profile.reportDings end,
				set = function()
				 QuestAgent.db.profile.reportDings = not QuestAgent.db.profile.reportDings
				end,
			 order = 11,
			},
			includeQuestAgentInTitle = {
				type = "toggle",
				name = L["Include QuestAgent in title"],
				desc = L["includeQuestAgentInTitle"],
				get = function() return QuestAgent.db.profile.includeQuestAgentInTitle end,
				set = function()
				 QuestAgent.db.profile.includeQuestAgentInTitle = not QuestAgent.db.profile.includeQuestAgentInTitle
				 QuestAgent:UpdateDisplay()
				end,
			 order = 11,
			},
			showPlayerExp = {
				type = "toggle",
				name = L["Show player experience"],
				desc = L["showPlayerExp"],
				get = function() return QuestAgent.db.profile.showPlayerExp end,
				set = function()
				 QuestAgent.db.profile.showPlayerExp = not QuestAgent.db.profile.showPlayerExp
				 QuestAgent:UpdateDisplay()
				end,
			 order = 11,
			},
			showAchievements = {
				type = "toggle",
				name = L["Show achievement progress"],
				desc = L["showAchievements"],
				get = function() return QuestAgent.db.profile.showPlayerExp end,
				set = function()
				 QuestAgent.db.profile.showAchievements = not QuestAgent.db.profile.showAchievements
				 QuestAgent:UpdateDisplay()
				end,
			 order = 11,
			},
		 spacer2 = {
			 type = "header",
			 name = " ",
			 order = 20,
		 },
		 tooltipOptions = {
		  name = L["Tooltip"],
		  type = "group",
		  desc = L["tooltipOptions"],
		  order = 21,
		  args = {
			  colorByDifficulty = {
				  type = "toggle",
				  name = L["Color quests by difficulty"],
				  desc = L["colorByDifficulty"],
				  get = function() return QuestAgent.db.profile.colorByDifficulty end,
				  set = function()
				   QuestAgent.db.profile.colorByDifficulty = not QuestAgent.db.profile.colorByDifficulty
				   QuestAgent:UpdateDisplay()
				  end,
			   order = 11,
			  },
			  detachTooltip = {
				  type = "toggle",
				  name = L["Detach Tooltip"],
				  desc = L["detachTooltip"],
				  get = function()
				   if tipFrame then
				    return not tablet:IsAttached(tipFrame)
				   else
				    return false
				   end
				  end,
				  set = function()
       if tipFrame ~= nil then
        if tablet:IsAttached(tipFrame) then
         tablet:Open(tipFrame)
         tablet:Detach(tipFrame)
         self.db.profile.detachTooltip = true
        else
         tablet:Attach(tipFrame)
         self.db.profile.detachTooltip = false
        end
       end
				   QuestAgent:UpdateDisplay()
				  end,
			   order = 11,
			  },
			  sortAlpha = {
				  type = "toggle",
				  name = L["Sort quests alphabetically"],
				  desc = L["sortAlpha"],
				  get = function() return QuestAgent.db.profile.sortAlpha end,
				  set = function()
				   QuestAgent.db.profile.sortAlpha = not QuestAgent.db.profile.sortAlpha
				   QuestAgent:UpdateDisplay()
				  end,
			   order = 11,
			  },
	    spacer1 = {
		    type = "header",
		    name = " ",
		    order = 20,
	    },
			  tooltipScale = {
			   type = "range",
			   name = L["Tooltip Scale"],
			   desc = L["tooltipScale"],
			   min = 0.25,
			   max = 4.00,
			   isPercent = true,
			   get = function() return QuestAgent.db.profile.tooltipScale end,
			   set = function(v)
			    QuestAgent.db.profile.tooltipScale = v
			    QuestAgent:UpdateDisplay()
			   end,
			   order = 31,
			  },
			  tooltipOpacity = {
			   type = "range",
			   name = L["Tooltip Opacity"],
			   desc = L["tooltipOpacity"],
			   min = 0.00,
			   max = 1.00,
			   isPercent = true,
			   get = function() return QuestAgent.db.profile.tooltipOpacity end,
			   set = function(v)
			    QuestAgent.db.profile.tooltipOpacity = v
			    QuestAgent:UpdateDisplay()
			   end,
			   order = 31,
			  },
			  showHint = {
				  type = "toggle",
			   name = L["Show hint"],
			   desc = L["showHint"],
				  get = function() return QuestAgent.db.profile.showHint end,
				  set = function()
				   QuestAgent.db.profile.showHint = not QuestAgent.db.profile.showHint
				   QuestAgent:UpdateDisplay()
				  end,
			   order = 31,
			  },
			 },
			},
		 filters = {
		  name = L["Filters"],
		  type = "group",
		  desc = L["filters"],
		  order = 31,
		  args = {
			  filterMyQuests = {
				  type = "toggle",
				  name = L["Only show quests that you also have"],
				  desc = L["filterMyQuests"],
				  get = function() return QuestAgent.db.profile.filterMyQuests end,
				  set = function()
				   QuestAgent.db.profile.filterMyQuests = not QuestAgent.db.profile.filterMyQuests
				   if QuestAgent.db.profile.filterMyQuests then QuestAgent.db.profile.filterNotMyQuests = false end
				   QuestAgent:UpdateDisplay()
				  end,
			  },
			  filterNotMyQuests = {
				  type = "toggle",
				  name = L["Only show quests that you don't have"],
				  desc = L["filterNotMyQuests"],
				  get = function() return QuestAgent.db.profile.filterNotMyQuests end,
				  set = function()
				   QuestAgent.db.profile.filterNotMyQuests = not QuestAgent.db.profile.filterNotMyQuests
				   if QuestAgent.db.profile.filterNotMyQuests then QuestAgent.db.profile.filterMyQuests = false end
				   QuestAgent:UpdateDisplay()
				  end,
			  },
			  filterZone = {
				  type = "toggle",
				  name = L["Only show quests in the current zone"],
				  desc = L["filterZone"],
				  get = function() return QuestAgent.db.profile.filterZone end,
				  set = function()
				   QuestAgent.db.profile.filterZone = not QuestAgent.db.profile.filterZone
				   QuestAgent:UpdateDisplay()
				  end,
			  },
			  filterComplete = {
				  type = "toggle",
				  name = L["Hide completed quests"],
				  desc = L["filterComplete"],
				  get = function() return QuestAgent.db.profile.filterComplete end,
				  set = function()
				   QuestAgent.db.profile.filterComplete = not QuestAgent.db.profile.filterComplete
				   QuestAgent:UpdateDisplay()
				  end,
			  },
			  filterWatched = {
				  type = "toggle",
				  name = L["Only show quests you are watching"],
				  desc = L["filterWatched"],
				  get = function() return QuestAgent.db.profile.filterWatched end,
				  set = function()
				   QuestAgent.db.profile.filterWatched = not QuestAgent.db.profile.filterWatched
				   QuestAgent:UpdateDisplay()
				  end,
			  },
		  },
		 },
		 spacer4 = {
			 type = "header",
			 name = " ",
			 order = 40,
		 },
		 header2 = {
			 type = "header",
			 name = L["Actions"],
			 order = 41,
		 },
			diffsync = {
				type = "execute",
				name = L["Perform differential sync now"],
				desc = L["diffsync"],
				buttonText = "Perform differential sync now",
				func = function()
				 self:Print(L["Synchronizing quests with party (differential only)..."])
				 self:SendDiffSync(self:GetMyQuests(), "PARTY")
     self:SendCommMessage(commPrefix.."RDiffSync", "", "PARTY")
				end,
			 order = 42,
			},
			fullsync = {
				type = "execute",
				name = L["Perform full sync now"],
				desc = L["fullsync"],
				buttonText = "Perform full sync now",
				func = function()
				 self:Print(L["Synchronizing quests with party..."])
     self:DoFullUpdate()
     self:SendCommMessage(commPrefix.."RFullUpd", "", "PARTY")
				end,
			 order = 42,
			},
		 spacer5 = {
			 type = "header",
			 name = " ",
			 order = 9999,
		 },
		}
	}
	
	-- Initialize database
 self.db = LibStub("AceDB-3.0"):New("QuestAgentDB", {
  profile = {
	  reportLogChanges = true,
	  colorByDifficulty = true,
	  showPetExp = true,
	  reportDings = true,
	  includeQuestAgentInTitle = false,
	  showPlayerExp = true,
	  filterMyQuests = false,
	  filterNotMyQuests = false,
	  filterZone = false,
	  filterComplete = false,
	  tooltipScale = 1.00,
	  tooltipOpacity = 0.65,
	  detachedData = {},
	  sortAlpha = false,
	  detachTooltip = false,
	  filterWatched = false,
	  showHint = true,
	  showAchievements = true,
   minimap = {},
	 }
 }, "Default")
 

end


function QuestAgent:OnEnable()
 self:ScheduleTimer("DelayedInit", 5)

 -- Initialize LDB object
	brokerPlugin = ldb:NewDataObject("QuestAgent", {
	 type = "data source",
	 icon = [[Interface\AddOns\QuestAgent\Art\icon]],
	 label = "QuestAgent",
	 tocname = "QuestAgent",
	})
	
	brokerPlugin.OnClick = function(frame, button)
		if button == "RightButton" then
	  if not dewdrop or not menu or not frame then return end
	  if not dewdrop:IsRegistered(frame) then
		  dewdrop:InjectAceOptionsTable(frame, menu)
		  dewdrop:Register(frame,
			  "children", menu,
			  "point", function(frame)
				  local x, y = frame:GetCenter()
				  local leftRight
				  if x < GetScreenWidth() / 2 then
					  leftRight = "LEFT"
				  else
					  leftRight = "RIGHT"
				  end
				  if y < GetScreenHeight() / 2 then
					  return "BOTTOM" .. leftRight, "TOP" .. leftRight
				  else
					  return "TOP" .. leftRight, "BOTTOM" .. leftRight
				  end
			  end,
			  "dontHook", true
		  )
	  end
	  dewdrop:Open(frame)
		end
	end
	
 brokerPlugin.OnEnter = function(frame)
  if not frame then return end
  if not tabletRegistered then
   tablet:Register(frame,
    "children", function()
     tablet:SetTitle("QuestAgent")
     self:UpdateTooltip()
    end,
    "detachedData", self.db and self.db.profile.detachedTooltip or {},
    "point", function(frame)
     if frame:GetTop() > GetScreenHeight() / 2 then
      local x = frame:GetCenter()
      if x < GetScreenWidth() / 3 then
       return "TOPLEFT", "BOTTOMLEFT"
      elseif x < GetScreenWidth() * 2 / 3 then
       return "TOP", "BOTTOM"
      else
       return "TOPRIGHT", "BOTTOMRIGHT"
      end
     else
      local x = frame:GetCenter()
      if x < GetScreenWidth() / 3 then
       return "BOTTOMLEFT", "TOPLEFT"
      elseif x < GetScreenWidth() * 2 / 3 then
       return "BOTTOM", "TOP"
      else
       return "BOTTOMRIGHT", "TOPRIGHT"
      end
     end
    end,
    "clickable", true,
    "hideWhenEmpty", true
   )	
   tabletRegistered = true
  end
  tipFrame = frame

  --if self.db.profile.detachTooltip then
   --tablet:Open(tipFrame)
   --tablet:Detach(tipFrame)
  --end

  self:UpdateDisplay()

 end
	
end


function QuestAgent:DelayedInit()
 self:RegisterEvent("PLAYER_XP_UPDATE", "SendExperienceUpdate")
 self:RegisterEvent("UNIT_PET", "SendExperienceUpdate")
 self:RegisterEvent("UNIT_PET_EXPERIENCE", "SendExperienceUpdate")
 self:RegisterEvent("QUEST_LOG_UPDATE", "DoPartialUpdate")
 self:RegisterEvent("QUEST_ACCEPTED", "DoPartialUpdate")
 self:RegisterEvent("UNIT_QUEST_LOG_CHANGED", "DoPartialUpdate")
 self:RegisterEvent("QUEST_WATCH_UPDATE", "DoPartialUpdate")
 self:RegisterEvent("QUEST_POI_UPDATE", "DoPartialUpdate")
 self:RegisterEvent("GROUP_ROSTER_UPDATE", "PartyMembersChanged")
 self:RegisterEvent("PARTY_MEMBER_ENABLE", "PartyMembersChanged")
 self:RegisterEvent("PARTY_MEMBER_DISABLE", "PartyMembersChanged")
 self:RegisterEvent("PLAYER_LEVEL_UP", "UpdateDisplay")
 self:RegisterEvent("ZONE_CHANGED_NEW_AREA", "UpdateDisplay")
	self:RegisterEvent("ZONE_CHANGED_INDOORS", "UpdateDisplay")
 self:RegisterEvent("ZONE_CHANGED", "UpdateDisplay")
 self:RegisterEvent("MINIMAP_ZONE_CHANGED", "UpdateDisplay")
 self:RegisterComm(commPrefix.."FullData", "OnCommReceived_FullData")
 self:RegisterComm(commPrefix.."RFullUpd", "OnCommReceived_FullUpdateRequest")
 self:RegisterComm(commPrefix.."RLat", "OnCommReceived_LatencyRequest")
 self:RegisterComm(commPrefix.."Lat", "OnCommReceived_LatencyReply")
 self:RegisterComm(commPrefix.."Exp", "OnCommReceived_ExperienceUpdate")
 self:RegisterComm(commPrefix.."ExpPet", "OnCommReceived_ExperienceUpdatePet")
 self:RegisterComm(commPrefix.."SyncStart", "OnCommReceived_SyncStartPlayer")
 self:RegisterComm(commPrefix.."QDel", "OnCommReceived_DeleteQuest")
 self:RegisterComm(commPrefix.."QUpd", "OnCommReceived_UpdateQuest")
 self:RegisterComm(commPrefix.."DiffSync", "OnCommReceived_DiffSync")
 self:RegisterComm(commPrefix.."RDiffSync", "OnCommReceived_RequestDiffSync")
 self:RegisterComm(commPrefix.."ROneUpd", "OnCommReceived_RequestSingleUpdate")
 self:RegisterComm(commPrefix.."RPush", "OnCommReceived_RequestPush")
 self:RegisterComm(commPrefix.."Ach", "OnCommReceived_AchievementData")
 self:RegisterComm(commPrefix.."Ver", "OnCommReceived_Version")
	self:RawHook("AddQuestWatch", true)
	self:RawHook("RemoveQuestWatch", true)
 self:HookQuestLog()
 self:DoFullUpdate()
 self:UpdateDisplay()
 self:SendCommMessage(commPrefix.."Ver", self:Serialize(self.version), "PARTY")
 if GetNumPartyMembers() > 0 then
  self:SendCommMessage(commPrefix.."RFullUpd", "", "PARTY")
 end
 --dbicon:Register("QuestAgent", brokerPlugin, self.db.profile.minimap)
 --self:AddPlayer(self:GetMyQuests())
end

StaticPopupDialogs["COPY_QUESTAGENT_URL"] = {
	text = L["Press CTRL-C to copy the URL to the clipboard"],
	button1 = L["Close"],
	hasEditBox = 1,	
	button2 = "",
	
	OnShow = function(self)
		self.editBox:SetText(L["url"]);
		self.editBox:SetFocus();
		self.editBox:HighlightText();
		_G[self:GetName().."Button2"]:Hide();
		_G[self:GetName().."Button1"]:ClearAllPoints();
		_G[self:GetName().."Button1"]:SetPoint("TOP", self.editBox, "BOTTOM", 0, -8);		
	end,
	OnHide = function(self)
		if (ChatFrameEditBox:IsVisible()) then
			ChatFrameEditBox:SetFocus();
		end
		self.editBox:SetText("");
		_G[self:GetName().."Button2"]:Show();
	end,
	EditBoxOnEnterPressed = function(self)
		self:GetParent():Hide()
	end,
	EditBoxOnEscapePressed = function(self)
		self:GetParent():Hide()
	end,
	timeout = 0,
	exclusive = 1,
	whileDead = 1,
	hideOnEscape = 1
};



--[[-- MISC HELPERS ------------------------------------------------------------------------------------------------------------]]

function QuestAgent:Debug(...)
 if debug then self:Print(...) end
end

local encode_translate = {
 [255] = "\255\001",
 [0] = "\255\002"
}

local function encode_helper(char)
 return encode_translate[char:byte()]
end

local decode_translate = {
 ["\001"] = "\255",
 ["\002"] = "\000"
}

local function decode_helper(text)
 return decode_translate[text]
end

function QuestAgent:Encode(data)
 return data:gsub("([\255%z])", encode_helper)
end

function QuestAgent:Decode(data)
 return data:gsub("\255([\001\002])", decode_helper)
end

function QuestAgent:Compress(data)
 return self:Encode(compress:Compress(data))
end

function QuestAgent:Decompress(data)
 return compress:Decompress(self:Decode(data))
end



--[[-- ACTIONS -----------------------------------------------------------------------------------------------------------------]]

function QuestAgent:PartyMembersChanged()
 self:SendCommMessage(commPrefix.."Ver", self:Serialize(self.version), "PARTY")
 self:DoFullUpdate()
end


function QuestAgent:SendAchievementUpdate()
 local s = self:Serialize(self:GetMyAchievements())
 if not s then
  self:Debug("[SendAchievementUpdate] Error serializing achievement information")
  return
 end
 self:SendCommMessage(commPrefix.."Ach", s, "PARTY")
 self:Debug("Sent achievement update. (Size: "..strlen(s)..")")
end


function QuestAgent:SendExperienceUpdate()
 local s = self:Serialize({ UnitName("player"), UnitXP("player"), UnitXPMax("player"), UnitLevel("player") })
 if not s then
  self:Debug("[SendExperienceUpdate] Error serializing player experience information")
  return
 end
 self:SendCommMessage(commPrefix.."Exp", s, "PARTY")
 self:Debug("Sent player experience update. (Size: "..strlen(s)..")")
 
 local pn = UnitName("pet")
 if pn ~= nil then
  local currXP, nextXP = GetPetExperience();
  s = self:Serialize({  UnitName("player"), pn, currXP, nextXP, UnitLevel("pet") })
  if not s then
   self:Debug("[SendExperienceUpdate] Error serializing pet experience information")
   return
  end
  self:SendCommMessage(commPrefix.."ExpPet", s, "PARTY")
  self:Debug("Sent pet experience update. (Size: "..strlen(s)..")")
  if pn == "Unknown" then
   self:Debug("Pet name unknown. Scheduling another update in 3 seconds.")
   self:ScheduleTimer(function()
    self:SendExperienceUpdate()
   end, 3)
  end
 end
end


function QuestAgent:SendFullUpdate(type, destination, overrideThrottle)
 if GetTime() - last < throttleTime and not overrideThrottle then
  if not inThrottle then
   inThrottle = true
   self:ScheduleTimer(function()
    self:SendFullUpdate(type, destination)
   end, throttleTime + 0.1)
   self:Debug("Tried to update before throttle expired. Setting up delayed update.")
  else
   self:Debug("Tried to update before throttle expired. Delayed update already set.")
  end
 else
  local o = self:GetMyQuests()
  if not o then
   self:Print("[SendFullUpdate] Error getting quest information")
   return
  end
  local s = self:Serialize(o)
  if not s then
   self:Print("[SendFullUpdate] Error serializing quest information")
   return
  end
  local c = self:Compress(s)
  if not c then
   self:Print("[SendFullUpdate] Error compressing quest information")
   return
  end
  if not type then
   self:Print("[SendFullUpdate] Error: Type not specified")
   return
  end
  if type == "WHISPER" and not destination then
   self:Print("[SendFullUpdate] Error: Destination not specified")
   return
  end
  self:SendCommMessage(commPrefix.."SyncStart", self:Serialize(o.name), type, destination)
  self:Debug("Sent sync start message for player "..o.name)
  self:SendCommMessage(commPrefix.."FullData", c, type, destination)
  self:Debug("Sent message packet. (Normal: "..strlen(s).." / Compressed: "..strlen(c)..")")
  last = GetTime()
  self:SendAchievementUpdate()
  inThrottle = false
 end
end


function QuestAgent:SendSingleUpdate(q)
 if not q then
  self:Debug("[SendSingleUpdate] No quest supplied")
  return
 end
 local s = self:Serialize(q)
 if not s then
  self:Debug("[SendSingleUpdate] Error serializing quest")
  return
 end
 local c = self:Compress(s)
 if not c then
  self:Debug("[SendSingleUpdate] Error compressing quest")
  return
 end
 self:SendCommMessage(commPrefix.."QUpd", c, "PARTY")
 self:Debug("Sent single quest packet. (Normal: "..strlen(s).." / Compressed: "..strlen(c)..")")
end


function QuestAgent:ROneUpd(destination, questId)
 self:SendCommMessage(commPrefix.."ROneUpd", self:Serialize(questId), "WHISPER", destination)
end


function QuestAgent:SendDiffSync(o, msgtype, destination, overrideThrottle)
 if GetTime() - lastDiff < throttleTimeDiff and not overrideThrottle then
  if not inThrottleDiff then
   inThrottleDiff = true
   self:ScheduleTimer(function()
    self:SendDiffSync(o, msgtype, destination)
   end, throttleTimeDiff + 0.1)
   self:Debug("Diff: Tried to update before throttle expired. Setting up delayed update.")
  else
   self:Debug("Diff: Tried to update before throttle expired. Delayed update already set.")
  end
 else
  if not msgtype then
   self:Debug("[SendDiffSync] Error: Type not specified")
   return
  end
  if msgtype == "WHISPER" and not destination then
   self:Debug("[SendDiffSync] Error: Destination not specified")
   return
  end
  if type(o.quests) == "table" then
   local d = {}
   local k, v
   for k, v in pairs(o.quests) do
    table.insert(d, { v.questId, v.objectiveList, v.isComplete })
   end
   local s = self:Serialize(d)
   if not s then
    self:Debug("[SendDiffSync] Error serializing information")
    return
   end
   local c = self:Compress(s)
   if not c then
    self:Debug("[SendDiffSync] Error compressing information")
    return
   end
   self:Debug("Sent differential synchronization packet (Normal: "..strlen(s).." / Compressed: "..strlen(c)..")")
   self:SendCommMessage(commPrefix.."DiffSync", c, msgtype, destination)
   self:SendAchievementUpdate()
  end
  lastDiff = GetTime()
  inThrottleDiff = false
 end
end


function QuestAgent:DoFullUpdate()
 self:SendExperienceUpdate()
 if GetNumPartyMembers() > 0 then self:SendFullUpdate("PARTY") end
end


function QuestAgent:DoPartialUpdate()
 self:SendDiffSync(self:GetMyQuests(), "PARTY")
end


function QuestAgent:ReportQuestDelete(player, questId)
 if QuestAgent.db.profile.reportLogChanges == true then
  local k, v
  for k, v in pairs(log) do
   if v.name == player then
    local kk, vv
    if type(v.quests) == "table" then
     for kk, vv in pairs(v.quests) do
      if vv.questId == questId then
       self:Debug("Quest "..vv.questId.." isComplete: "..(isComplete or "nil"))
       if vv.savedIsComplete == 1 or vv.isComplete == 1 then
        self:Print(string.format(L["turned in"], player, vv.questLink))
       else
        self:Print(string.format(L["abandoned"], player, vv.questLink))
       end
      end
     end
    end
   end
  end
 end
end


function QuestAgent:ReportQuestAdd(player, questId, questLink)
 if QuestAgent.db.profile.reportLogChanges == true then
  local k, v, found
  found = false
  for k, v in pairs(log) do
   if v.name == player then
    local kk, vv
    if type(v.quests) == "table" then
     for kk, vv in pairs(v.quests) do
      if vv.questId == questId then
       found = true
      end
     end
    end
   end
  end
  if found == false then
   self:Print(string.format(L["accepted new quest"], player, questLink))
  end
 end
end


function QuestAgent:ReportQuestComplete(player, questId)
 if QuestAgent.db.profile.reportLogChanges == true then
  local k, v, found, questLink
  found = false
  for k, v in pairs(log) do
   if v.name == player then
    local kk, vv
    if type(v.quests) == "table" then
     for kk, vv in pairs(v.quests) do
      if vv.questId == questId then
       found = true
       questLink = vv.questLink
      end
     end
    end
   end
  end
  if found == true then
   self:Print(string.format(L["completed quest"], player, questLink))
  end
 end
end


function QuestAgent:QDel(player, questId)
 local k, v
 for k, v in pairs(log) do
  if v.name == player and type(v.quests) == "table" then
   local kk, vv
   for kk, vv in pairs(v.quests) do
    if vv.questId == questId then
     log[k].quests[kk] = nil
     log[k].lastUpdate = GetTime()
     self:Debug("Deleted quest "..questId.." for player "..player)
    end
   end
  end
 end
 log.deleteTrigger = 0
 log.deleteTrigger = nil
end


function QuestAgent:AddPlayer(o)
 if o == nil then return end
 if not o.name then return end
 self:Debug("Updated player "..o.name)
 local isCollapsed = self:RemovePlayer(o.name)
 o.isCollapsed = isCollapsed
 o.lastUpdate = GetTime()
 table.insert(log, o)
 self:UpdateDisplay()
end


function QuestAgent:RemovePlayer(name)
 local k, v, isCollapsed
 isCollapsed = false
 for k, v in pairs(log) do
  if v.name == name then
   isCollapsed = v.isCollapsed
   log[k] = nil
  end
 end
 log.deleteTrigger = 0
 log.deleteTrigger = nil
 return isCollapsed
end



--[[-- COMMUNICATION HANDLERS --------------------------------------------------------------------------------------------------]]

function QuestAgent:OnCommReceived_AchievementData(prefix, message, distribution, sender)
 if sender == (UnitName("player")) then return end
 local success, o = self:Deserialize(message)
 if success == false then
  self:Debug("[OnCommReceived_AchievementData] Failed to deserialize message.")
  return
 else
  self:Debug("Received achievement update from "..sender)
  ach[sender] = o
  self:UpdateDisplay()
 end
end


function QuestAgent:OnCommReceived_Version(prefix, message, distribution, sender)
 if sender == (UnitName("player")) then return end
 local success, o = self:Deserialize(message)
 if success == false then
  self:Debug("[OnCommReceived_Version] Failed to deserialize message.")
  return
 else
  self:Debug(sender.." is running QuestAgent r"..o)
  ver[sender] = o
  self:UpdateDisplay()
  if o > self.version and versionWarned == false then
   versionWarned = true
			QuestAgentUpdateDialogText:SetText(string.format(L["update"], sender, o));
			QuestAgentUpdateDialog:Show();
  end
 end
end


function QuestAgent:OnCommReceived_RequestPush(prefix, message, distribution, sender)
 if sender == (UnitName("player")) then return end
 local success, o = self:Deserialize(message)
 if success == false then
  self:Debug("[OnCommReceived_RequestPush] Failed to deserialize message.")
  return
 else
  local l = self:GetMyQuests()
  local k, v
  for k, v in pairs(l.quests) do
   if v.questId == o then
    self:Debug("Received push request for quest "..v.questId)
    self:PushQuest(o, sender)
   end
  end
 end
end


function QuestAgent:OnCommReceived_RequestDiffSync(prefix, message, distribution, sender)
 if sender == (UnitName("player")) then return end
 self:Debug("Received differential sync request from "..sender)
 self:SendDiffSync(self:GetMyQuests(), "WHISPER", sender)
end


function QuestAgent:OnCommReceived_RequestSingleUpdate(prefix, message, distribution, sender)
 if sender == (UnitName("player")) then return end
 local success, o = self:Deserialize(message)
 if success == false then
  self:Debug("[OnCommReceived_RequestSingleUpdate] Failed to deserialize message.")
  return
 else
  local l = self:GetMyQuests()
  local k, v
  for k, v in pairs(l.quests) do
   if v.questId == o then
    self:Debug("Received single update request. Sending update to party for quest "..v.questId..".")
    self:SendSingleUpdate(v)
   end
  end
 end
end


function QuestAgent:OnCommReceived_DiffSync(prefix, message, distribution, sender)
 if sender == (UnitName("player")) then return end

 local success, o = self:Deserialize(self:Decompress(message))
 if success == false then
  self:Debug("[OnCommReceived_DiffSync] Failed to deserialize message.")
  return
 else
  if o == nil then return end
  if sender == UnitName("player") then return end
  self:Debug("Received differential synchronization for player "..sender)
  
  local a, b, l
  for a, b in pairs(log) do
   if b.name == sender then
    l = b
    break
   end
  end
  if l == nil then
   self:Debug("Could not locate player "..sender.." for differential update")
   return
  end
  
  if l ~= nil and type(l.quests) == "table" then
  
   if l.inSync == true then
    self:Debug("Differential sync not accepted, player "..sender.." was already in the middle of a full synchronization.")
    return
   end
  
   -- Check for quests to be deleted
   for k, v in pairs(l.quests) do
    local kk, vv, found
    found = false
    for kk, vv in pairs(o) do
     if v.questId == vv[1] then found = true end
    end
    if found == false then
     self:Debug("Quest "..v.questId.." removed during differential sync")
     self:ReportQuestDelete(sender, v.questId)
     self:QDel(sender, v.questId)
     if QuestAgent_QuestLogFrame:IsShown() and QuestAgent.currentQuest == v.questId and QuestAgent.currentPlayer == sender then
      self:Debug("Currently shown quest removed. Closing quest log...")
      QuestAgent_ToggleQuestLog()
     end
    end
   end
  
   -- Check for quests to be added
   for k, v in pairs(o) do
    local kk, vv, found
    found = false
    for kk, vv in pairs(l.quests) do
     if v[1] == vv.questId then found = true end
    end
    if found == false then
     self:Debug("New quest found during differential sync. Requesting update for quest "..v[1]..".")
     self:ROneUpd(sender, v[1])
    end
   end
  
   -- Update all quest objectives
   local updated = 0
   for a, b in pairs(log) do
    if b.name == sender then
     for k, v in pairs(o) do
      local kk, vv
      for kk, vv in pairs(b.quests) do
       if v[1] == vv.questId then
        updated = updated + 1
        log[a].quests[kk].objectiveList = v[2]
        if v[3] == 1 then log[a].quests[kk].savedIsComplete = 1 end
								if not log[a].quests[kk].isComplete and v[3] then
									self:ReportQuestComplete(sender, vv.questId)
								end
        log[a].quests[kk].isComplete = v[3]
        if QuestAgent_QuestLogFrame:IsShown() and QuestAgent.currentQuest == v[1] and QuestAgent.currentPlayer == sender then
         self:Debug("Currently shown quest changed. Rerendering quest log...")
         self:ShowQuest(vv, sender)
        end
       end
      end
     end
    end
   end
   self:Debug("Updated objectives for "..updated.." quest(s) for player "..sender.." during differential sync")
   
  end
  
  self:UpdateDisplay()
 end
end


function QuestAgent:OnCommReceived_UpdateQuest(prefix, message, distribution, sender)
 local success, o = self:Deserialize(self:Decompress(message))
 if success == false then
  self:Debug("[OnCommReceived_UpdateQuest] Failed to deserialize message.")
  return
 else
  if o == nil then return end
  if sender == (UnitName("player")) then return end
  self:Debug("Received quest update from player "..sender.." for quest "..o.questId)
  self:ReportQuestAdd(sender, o.questId, o.questLink)
  self:QDel(sender, o.questId)
  local k, v
  for k, v in pairs(log) do
   if v.name == sender and type(v.quests) == "table" then
    table.insert(log[k].quests, o)
    log[k].lastUpdate = GetTime()
    self:Debug("Added quest "..o.questId.." for player "..sender)
   end
  end
  self:UpdateDisplay()
  if QuestAgent_QuestLogFrame:IsShown() and QuestAgent.currentQuest == o.questId and QuestAgent.currentPlayer == sender then
   self:Debug("Currently shown quest changed. Rerendering quest log...")
   self:ShowQuest(o, sender)
  end
 end
end


function QuestAgent:OnCommReceived_DeleteQuest(prefix, message, distribution, sender)
 local success, o = self:Deserialize(message)
 if success == false then
  self:Debug("[OnCommReceived_DeleteQuest] Failed to deserialize message.")
  return
 else
  if o == nil then return end
  if sender == (UnitName("player")) then return end
  self:Debug("Received quest delete command from player "..sender.." for quest "..o)
  self:ReportQuestDelete(sender, o)
  self:QDel(sender, o)
  self:UpdateDisplay()
 end
end


function QuestAgent:OnCommReceived_SyncStartPlayer(prefix, message, distribution, sender)
 local success, o = self:Deserialize(message)
 if success == false then
  self:Debug("[OnCommReceived_SyncStartPlayer] Failed to deserialize message.")
  return
 else
  if o == nil then return end
  if o == (UnitName("player")) then return end
  if o ~= sender then o = sender end
  self:Debug("Received sync start for player "..o)
  local isCollapsed = self:RemovePlayer(o)
  table.insert(log, { name = o, isCollapsed = isCollapsed, quests = {}, inSync = true, syncStart = GetTime() })
  self:UpdateDisplay()
 end
end


function QuestAgent:OnCommReceived_FullData(prefix, message, distribution, sender)
 local success, o = self:Deserialize(self:Decompress(message))
 if success == false then
  self:Debug("[OnCommReceived_FullData] Failed to deserialize full data update. Error message: "..o.."\"")
  return
 else
  if o == nil then return end
  if not o.name then return end
  if o.name == (UnitName("player")) then return end
  if o.name ~= sender then o.name = sender end
  self:AddPlayer(o)
  if o.timeCreated then
   self:Debug("Requesting latency from "..sender)
   self:SendCommMessage(commPrefix.."RLat", self:Serialize(o.timeCreated), "WHISPER", sender)
  end
  self:Debug("Requesting differential sync from "..sender)
  self:SendCommMessage(commPrefix.."RDiffSync", "", "WHISPER", sender)
 end
end


function QuestAgent:OnCommReceived_ExperienceUpdatePet(prefix, message, distribution, sender)
 local success, o = self:Deserialize(message)
 if success == false then
  self:Debug("[OnCommReceived_ExperienceUpdatePet] Failed to deserialize pet experience update.")
  return
 else
  if o == nil then return end
  if not o[1] then return end
  if o[1] == (UnitName("player")) or sender == (UnitName("player")) then return end
  if o[1] ~= sender then o[1] = sender end
  if QuestAgent.db.profile.reportDings and QuestAgent.db.profile.showPetExp and exp[o[1]] and exp[o[1]][2] == o[2] and exp[o[1]][5] ~= o[5] then
   self:Print(string.format(L["pet level gain"], o[1], o[2], o[5]))
  end
  pet[o[1]] = o
  if o[3] == 0 and o[4] == 0 then pet[o[1]] = nil end
  self:UpdateDisplay()
  self:Debug("Received pet experience update from "..o[1].." for pet "..o[2]..". "..o[5]..": "..o[3].." / "..o[4])
 end
end


function QuestAgent:OnCommReceived_ExperienceUpdate(prefix, message, distribution, sender)
 local success, o = self:Deserialize(message)
 if success == false then
  self:Debug("[OnCommReceived_ExperienceUpdate] Failed to deserialize experience update.")
  return
 else
  if o == nil then return end
  if not o[1] then return end
  if o[1] == (UnitName("player")) or sender == (UnitName("player")) then return end
  if o[1] ~= sender then o[1] = sender end
  if QuestAgent.db.profile.reportDings and exp[o[1]] and exp[o[1]][4] ~= o[4] then
   self:Print(string.format(L["level gain"], o[1], o[4]))
  end
  exp[o[1]] = o
  self:UpdateDisplay()
  self:Debug("Received experience update from "..o[1]..". "..o[4]..": "..o[2].." / "..o[3])
 end
end


function QuestAgent:OnCommReceived_FullUpdateRequest(prefix, message, distribution, sender)
 if sender == (UnitName("player")) then return end
 self:Debug("Received an update request from "..sender)
 self:SendExperienceUpdate()
 self:SendFullUpdate("WHISPER", sender, true)
end


function QuestAgent:OnCommReceived_LatencyRequest(prefix, message, distribution, sender)
 local success, o = self:Deserialize(message)
 if success then
  self:Debug("Received a latency request from "..sender)
  self:SendCommMessage(commPrefix.."Lat", self:Serialize(GetTime() - o), "WHISPER", sender)
 else
  self:Debug("Received an invalid latency request from "..sender)
 end
end


function QuestAgent:OnCommReceived_LatencyReply(prefix, message, distribution, sender)
 local success, o = self:Deserialize(message)
 if success then
  local k, v
  for k, v in pairs(log) do
   if v.name == sender then
    v.latency = o
   end
  end
  self:UpdateDisplay()
  self:Debug("Received a latency response from "..sender)
 else
  self:Debug("Received an invalid latency response from "..sender)
 end
end



--[[-- DISPLAY -----------------------------------------------------------------------------------------------------------------]]

function QuestAgent:UpdateDisplay()
 self:UpdateText()
 if tipFrame ~= nil then
  tablet:Refresh(tipFrame)
  self:UpdateDisplayProperties()
 end
end


function QuestAgent:UpdateDisplayProperties()
 if tipFrame ~= nil then
  tablet:SetFontSizePercent(tipFrame, QuestAgent.db.profile.tooltipScale)
  tablet:SetTransparency(tipFrame, QuestAgent.db.profile.tooltipOpacity)
 end
end


function QuestAgent:C(r, g, b)
 return string.format("|cff%02x%02x%02x", (r*255), (g*255), (b*255));
end


function QuestAgent:MakeQuestTag(level, isDaily, questTag, suggestedGroup, colorize)
 if colorize == nil then colorize = true end
 local modifier = ""
 if isDaily then modifier = L["y"] end
 if questTag == ELITE then modifier = modifier..L["+"]
 elseif questTag == LFG_TYPE_DUNGEON then modifier = modifier..L["d"]
 elseif questTag == DUNGEON_DIFFICULTY2 then modifier = modifier..L["d+"]
 elseif questTag == PVP then modifier = modifier..L["p"]
 elseif questTag == RAID then modifier = modifier..L["r"]
 elseif questTag == GROUP then modifier = modifier..L["g"]
 end
 if suggestedGroup and suggestedGroup > 0 then modifier = modifier..suggestedGroup end
 
 local tag = "["
 if colorize then tag = tag..WHT end
 tag = tag..level
 if colorize then tag = tag.."|r"..HIL end
 tag = tag..modifier
 if colorize then tag = tag.."|r" end
 tag = tag.."]"
 return tag
end


function QuestAgent:OthersOnQuest(name, questId)
 local k, v, kk, vv
 local count = 0
 -- Check others in your log
 for k, v in pairs(log) do
  if type(v.name) == "string" and type(v.quests) == "table" then
   if v.name ~= name and v.name ~= UnitName("player") then
    for kk, vv in pairs(v.quests) do
     if vv.questId == questId then
      count = count + 1
     end
    end
   end
  end
 end
 -- Check yourself
 if myquests then
  for k, v in pairs(myquests.quests) do
   if v.questId == questId then
    count = count + 1
   end
  end
 end
 if count == 0 then return "" else return count end
end


function QuestAgent:MeOnQuest(questId)
 if myquests then
  for k, v in pairs(myquests.quests) do
   if v.questId == questId then
    return true
   end
  end
 end
 return false
end


function QuestAgent:IsWatched(questId)
 if myquests then
  for k, v in pairs(myquests.quests) do
   if v.questId == questId and v.isWatched then
    return true
   end
  end
 end
 return false
end


function QuestAgent:GetLastUpdate(lastUpdate)
 if lastUpdate == nil then return "(None)" end
 if GetTime() - lastUpdate >= 1 then
  return "("..SecondsToTime(GetTime() - lastUpdate)..")"
 end
 return L["(Now)"]
end


function QuestAgent:FormatNumber(num)
 return string.format("%d", num)
end


function QuestAgent:FormatSafePercent(a, b)
 local v
 if b == 0 then v = 0 else v = (a / b) * 100 end
 return string.format("%.2f", v).."%"
end


function QuestAgent:UpdateText()
 local numPlayers = 0
 for _ in pairs(log) do numPlayers = numPlayers + 1 end
 local s = numPlayers.."/"
 local i = 0
 for k, v in pairs(log) do
  if type(v.name) == "string" and type(v.quests) == "table" then
   for _ in pairs(v.quests) do i = i + 1 end
  end
 end
 s = s..i
 if QuestAgent.db.profile.includeQuestAgentInTitle then s = "QuestAgent ("..s..")" end
 brokerPlugin.text = s
end


function QuestAgent:FiltersActive()
 return self.db.profile.filterMyQuests or self.db.profile.filterNotMyQuests or self.db.profile.filterZone or self.db.profile.filterComplete or self.db.profile.filterWatched
end


local function comparator(a, b)
 if QuestAgent.db.profile.sortAlpha then
  if a.zone == b.zone then
   return a.questTitle < b.questTitle
  else
   return a.zone < b.zone
  end
 else
  if a.zone == b.zone then
   if a.level == b.level then
    return a.questTitle < b.questTitle
   else
    return a.level < b.level
   end
  else
   return a.zone < b.zone
  end
 end
end


function QuestAgent:SortQuests(t)
 local nt = {}
 local i, j, n, min = 0, 0, 0, 0
 local k, v
 for k, v in pairs(t) do
  n = n + 1
  nt[n] = v
 end
 for i = 1, n, 1 do
	 min = i
	 for j = i + 1, n, 1 do
		 if comparator(nt[j], nt[min]) then min = j end
	 end
	 nt[i], nt[min] = nt[min], nt[i]
 end
 return nt
end


function QuestAgent:UpdateTooltip()
 local cat
 local o = log
 local lastZone = ""
 local added = 0
 
 local k, v
 for k, v in pairs(o) do
  if type(v.name) == "string" and type(v.quests) == "table" then
  
   v.quests = self:SortQuests(v.quests)
   added = added + 1
 
   local latency = ""
   if v.latency then latency = SecondsToTime(v.latency) end
   
   local numQuests = 0
   if type(v.quests == "table") then
    for x, y in pairs(v.quests) do numQuests = numQuests + 1 end
   end
   
   local filter = ""
   if self:FiltersActive() then filter = L["filter active"] end
   
   local version = "Unknown"
   if ver[v.name] then version = "r"..ver[v.name] end
   
   cat = tablet:AddCategory(
    "columns", 4,
    "hideBlankLine", true,
    "showWithoutChildren", true,
    "checked", true,
    "hasCheck", true,
    "checkIcon", v.isCollapsed and "Interface\\Buttons\\UI-PlusButton-Up" or "Interface\\Buttons\\UI-MinusButton-Up",
    
    "text", v.name.." "..self:GetLastUpdate(v.lastUpdate),
    "text2", "("..numQuests.." quests)"..filter,
    "text3", version,
    
    "func", "OnQuestClick",
    "arg1", self,
    "arg2", "header",
    "arg3", v
   )
   
   if not v.isCollapsed then
    if self.db.profile.showPlayerExp and exp[v.name] then
     cat:AddLine(
      "text", HIL.."["..WHT..exp[v.name][4].."|r]|r "..exp[v.name][2].." / "..exp[v.name][3],
      "text2", L["Remaining: "]..self:FormatNumber(exp[v.name][3] - exp[v.name][2]).." ("..self:FormatSafePercent((exp[v.name][3] - exp[v.name][2]), exp[v.name][3])..")",
      "text3", self:FormatSafePercent(exp[v.name][2], exp[v.name][3])
     )
    end
    if self.db.profile.showPetExp and pet[v.name] then
     cat:AddLine(
      "text", pet[v.name][2].." "..HIL.."["..WHT..pet[v.name][5].."|r]|r "..pet[v.name][3].." / "..pet[v.name][4],
      "text2", L["Remaining: "]..self:FormatNumber(pet[v.name][4] - pet[v.name][3]).." ("..self:FormatSafePercent((pet[v.name][4] - pet[v.name][3]), pet[v.name][4])..")",
      "text3", self:FormatSafePercent(pet[v.name][3], pet[v.name][4])
     )
    end
    
    if v.inSync == true then
    
     local s = L["Synchronizing"]
     if v.syncStart ~= nil then
      if GetTime() - v.syncStart >= 1 then
       s = s.." ("..SecondsToTime(GetTime() - v.syncStart)..")"
      end
     end
     s = s.."..."
     cat:AddLine("text", HIL..s.."|r")
     
    else

     local kk, vv
     for kk, vv in pairs(v.quests) do
     
      local status = ""
      if vv.isComplete == 1 then status = COMPLETE
      elseif vv.isComplete == -1 then status = FAILED
      end
      
      local tag = self:MakeQuestTag(vv.level, vv.isDaily, vv.questTag, vv.suggestedGroup)
           
      local difficultyColor = GetQuestDifficultyColor(vv.level)
      if self.db.profile.colorByDifficulty ~= true then difficultyColor = { r = 1, g = 1, b = 1 } end
      
      local zone = ""
      if lastZone ~= vv.zone then
       zone = vv.zone
       lastZone = zone
      end
      
      local zoneDisplay = zone
      if self.db.profile.showAchievements then
       if zone ~= "" then
        local a = ach[v.name]
        if a then
         local id = self:GetAchievementId(zone)
         if id and a[id] then
          zoneDisplay = zoneDisplay..WHT.." (|r"..HIL..tostring(a[id][1]).."|r"..WHT.."/|r"..HIL..tostring(a[id][2]).."|r"..WHT..")|r"
         end
        end
       end
      end
      
      local showQuest = true
      if self.db.profile.filterMyQuests and not self:MeOnQuest(vv.questId) then showQuest = false end
      if self.db.profile.filterNotMyQuests and self:MeOnQuest(vv.questId) then showQuest = false end
      if self.db.profile.filterComplete and (vv.isComplete or vv.savedIsComplete) then showQuest = false end
      if self.db.profile.filterZone and (vv.zone ~= GetZoneText() and vv.zone ~= GetRealZoneText()) then showQuest = false end
      if self.db.profile.filterWatched and not self:IsWatched(vv.questId) then showQuest = false end
      
      if showQuest then
       cat:AddLine(
        "text", zoneDisplay, "textR", 0, "textG", 1, "textB", 0, "justify", "LEFT",
        "text2", tag.." "..self:C(difficultyColor.r, difficultyColor.g, difficultyColor.b)..vv.questTitle.."|r", "justify2", "LEFT",
        "text3", status, "justify3", "LEFT",
        "text4", self:OthersOnQuest(v.name, vv.questId), "justify4", "LEFT",
        "func", "OnQuestClick",
        "arg1", self,
        "arg2", "quest",
        "arg3", vv,
        "arg4", k,
        "arg5", kk
       )
      end
      
     end
     
    end
    
   end
  end
 end
 
 if added <= 0 then
  cat = tablet:AddCategory("columns", 1)
  cat:AddLine(
   "text",
   L["Hint: Join a party. If they're using QuestAgent, you'll see their quests after 2-3 minutes."],
   "textR", 0, "textG", 1, "textB", 0,
   "wrap", true
  )
 else
  if self.db.profile.showHint then
   cat = tablet:AddCategory("columns", 1)
   cat:AddLine(
    "text",
    L["Hint: Click a quest to open the quest log"]..". "..
    L["Shift-Click a quest to link it into chat"]..". "..
    L["Ctrl-Click a player to remove them"]..".",
    "textR", 0, "textG", 1, "textB", 0,
    "wrap", true
   )
  end
 end
end


function QuestAgent:OnQuestClick(type, data, key1, key2)
 if type == "header" then
  if IsControlKeyDown() then
   self:RemovePlayer(data.name)
   self:UpdateDisplay()
  else
   data.isCollapsed = not data.isCollapsed
   self:UpdateDisplay()
  end
 else
  if IsShiftKeyDown() then
   ChatEdit_InsertLink(data.questLink);
  else
   self:ShowQuest(data, log[key1].name)
  end
 end
end



--[[-- ACHIEVEMENTS ------------------------------------------------------------------------------------------------------------]]

local hordeAchievements = {
-- Kalimdor
	["Ashenvale"] = 4976,
	["Azshara"] = 4927,
	--["Azuremyst Isle"] = ,
	--["Bloodmyst Isle"] = ,
	--["Darkshore"] = ,
	["Desolace"] = 4930,
	--["Durotar"] = ,
	["Dustwallow Marsh"] = 4978,
	["Felwood"] = 4931,
	["Feralas"] = 4979,
	--["Moonglade"] = ,
	--["Mulgore"] = ,
	["Northern Barrens"] = 4933,
	["Silithus"] = 4934,
	["Southern Barrens"] = 4981,
	["Stonetalon Mountains"] = 4980,
	["Tanaris"] = 4935,
	--["Teldrassil"] = ,
	["Thousand Needles"] = 4938,
	["Un'Goro Crater"] = 4939,
	["Winterspring"] = 4940,
-- Eastern Kingdoms
	["Arathi Highlands"] = 4896,
	["Badlands"] = 4900,
	["Blasted Lands"] = 4909,
	["Burning Steppes"] = 4901,
	["The Cape of Stranglethorn"] = 4905,
	--["Deadwind Pass"] = ,
	--["Dun Morogh"] = ,
	--["Duskwood"] = ,
	["Eastern Plaguelands"] = 4982,
	--["Elwynn Forest"] = ,
	["Ghostlands"] = 4908,
	["Hillsbrad Foothills"] = 4985,
	["The Hinterlands"] = 4897,
	--["Loch Modan"] = ,
	["Northern Stranglethorn"] = 4906,
	--["Redridge Mountains"] = ,
	["Searing Gorge"] = 4910,
	["Silverpine Forest"] = 4894,
	["Swamp of Sorrows"] = 4904,
	--["Tirisfal Glades"] = ,
	["Western Plaguelands"] = 4893,
	--["Westfall"] = ,
	--["Wetlands"] = ,
	--["Eversong Woods"] = ,
	--["Isle of Quel'Danas"] = ,
 -- Outland
	["Blade's Edge Mountains"] = 1193,
	["Zangarmarsh"] = 1190,
	["Netherstorm"] = 1194,
	["Hellfire Peninsula"] = 1271,
	["Terokkar Forest"] = 1272,
	["Shadowmoon Valley"] = 1195,
	["Nagrand"] = 1273,
 -- Northrend
	["Icecrown"] = 40,
	["Dragonblight"] = 1359,
	["Howling Fjord"] = 1356,
	["Borean Tundra"] = 1358,
	["Sholazar Basin"] = 39,
	["Zul'Drak"] = 36,
	["Grizzly Hills"] = 1357,
	["The Storm Peaks"] = 38,
	-- Fixups for subzones
	["Coldarra"] = 1358,
 -- Cataclysm
	["Mount Hyjal"] = 4870,
	["Uldum"] = 4872,
	["Twilight Highlands"] = 5501,
	["Vashj'ir"] = 4982,
	["Kelp'thar Forest"] = 4982,
	["Shimmering Expanse"] = 4982,
	["Abyssal Depths"] = 4982,
	["Deepholm"] = 4871,
}

local allianceAchievements = {
-- Kalimdor
	["Ashenvale"] = 4925,
	--["Azshara"] = ,
	--["Azuremyst Isle"] = ,
	["Bloodmyst Isle"] = 4926,
	["Darkshore"] = 4928,
	["Desolace"] = 4930,
	--["Durotar"] = ,
	["Dustwallow Marsh"] = 4929,
	["Felwood"] = 4931,
	["Feralas"] = 4932,
	--["Moonglade"] = ,
	--["Mulgore"] = ,
	--["Northern Barrens"] = ,
	["Silithus"] = 4934,
	["Southern Barrens"] = 4937,
	["Stonetalon Mountains"] = 4936,
	["Tanaris"] = 4935,
	--["Teldrassil"] = ,
	["Thousand Needles"] = 4938,
	["Un'Goro Crater"] = 4939,
	["Winterspring"] = 4940,
-- Eastern Kingdoms
	["Arathi Highlands"] = 4896,
	["Badlands"] = 4900,
	["Blasted Lands"] = 4909,
	["Burning Steppes"] = 4901,
	["The Cape of Stranglethorn"] = 4905,
	--["Deadwind Pass"] = ,
	--["Dun Morogh"] = ,
	["Duskwood"] = 4903,
	["Eastern Plaguelands"] = 4982,
	--["Elwynn Forest"] = ,
	--["Ghostlands"] = ,
	--["Hillsbrad Foothills"] = ,
	["The Hinterlands"] = 4897,
	["Loch Modan"] = 4899,
	["Northern Stranglethorn"] = 4906,
	["Redridge Mountains"] = 4902,
	["Searing Gorge"] = 4910,
	--["Silverpine Forest"] = ,
	["Swamp of Sorrows"] = 4904,
	--["Tirisfal Glades"] = ,
	["Western Plaguelands"] = 4893,
	["Westfall"] = 4903,
	["Wetlands"] = 4899,
	--["Eversong Woods"] = ,
	--["Isle of Quel'Danas"] = ,
 -- Outland
	["Blade's Edge Mountains"] = 1193,
	["Zangarmarsh"] = 1190,
	["Netherstorm"] = 1194,
	["Hellfire Peninsula"] = 1189,
	["Terokkar Forest"] = 1191,
	["Shadowmoon Valley"] = 1195,
	["Nagrand"] = 1192,
 -- Northrend
	["Icecrown"] = 40,
	["Dragonblight"] = 35,
	["Howling Fjord"] = 34,
	["Borean Tundra"] = 33,
	["Sholazar Basin"] = 39,
	["Zul'Drak"] = 36,
	["Grizzly Hills"] = 37,
	["The Storm Peaks"] = 38,
	-- Fixups for subzones
	["Coldarra"] = 33,
 -- Cataclysm
	["Mount Hyjal"] = 4870,
	["Uldum"] = 4872,
	["Twilight Highlands"] = 4873,
	["Vashj'ir"] = 4869,
	["Kelp'thar Forest"] = 4869,
	["Shimmering Expanse"] = 4869,
	["Abyssal Depths"] = 4869,
	["Deepholm"] = 4871,
}

function QuestAgent:GetAchievementTable()
 local englishFaction, localizedFaction = UnitFactionGroup("player")
 if englishFaction == "Horde" then
  return hordeAchievements
 else
  return allianceAchievements
 end
end

function QuestAgent:GetAchievementId(zone)
 return self:GetAchievementTable()[lbz[zone] or zone]
end

function QuestAgent:GetMyAchievementStatus(id)
 if not pcall(
  function()
   if id then
    local criteriaString, criteriaType, completed, quantity, reqQuantity, charName, flags, assetID, quantityString, unknown = GetAchievementCriteriaInfo(id, 1)
    return quantity, reqQuantity
   else
    return nil
   end
  end)
 then
  self:Print("Error loading achievement information for ID: "..id)
  return nil
 end
end

function QuestAgent:GetMyAchievements()
 local k, v
 local t = {}
 for k, v in pairs(self:GetAchievementTable()) do
  local a, b = self:GetMyAchievementStatus(v)
  if a and b then t[v] = { a, b } end
 end
 return t
end




--[[-- QUEST MANIPULATION ------------------------------------------------------------------------------------------------------]]

function QuestAgent:AddQuestWatch(...)
 self.hooks["AddQuestWatch"](...)
 myquests = self:GetMyQuests()
 self:UpdateDisplay()
end


function QuestAgent:RemoveQuestWatch(...)
 self.hooks["RemoveQuestWatch"](...)
 myquests = self:GetMyQuests()
 self:UpdateDisplay()
end


function QuestAgent:PushQuest(questId, sender)
 local originalEntry = GetQuestLogSelection()
 local i = 0
 if sender == nil then sender = "<Unknown>" end
 while GetQuestLogTitle(i + 1) ~= nil do
  i = i + 1
  local questTitle, level, questTag, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily = GetQuestLogTitle(i)
  if not isHeader then
   SelectQuestLogEntry(i)
   if tonumber(GetQuestLink(i):match(":(%d+):")) == questId then
    QuestLogPushQuest()
   end
  end
 end
 SelectQuestLogEntry(originalEntry)
end


function QuestAgent:GetMyQuests()
 local originalEntry = GetQuestLogSelection()
 local o = { name = UnitName("player"), quests = {}, isCollapsed = false, timeCreated = GetTime() }
 local i = 0
 local zone = ""
 while GetQuestLogTitle(i + 1) ~= nil do
  i = i + 1
  local questTitle, level, questTag, suggestedGroup, isHeader, isCollapsed, isComplete, isDaily = GetQuestLogTitle(i)
  if not isHeader then

   SelectQuestLogEntry(i)
   local questDescription, questObjectives = GetQuestLogQuestText()
   
   local numObjectives = GetNumQuestLeaderBoards()
   local objectiveList = {}
	  for c = 1, numObjectives, 1 do
		  local text, type, finished = GetQuestLogLeaderBoard(c)
		  table.insert(objectiveList, {
		   text = text,
		   type = type,
		   finished = finished,
		  })
		 end
		 
		 local numChoices = GetNumQuestLogChoices()
		 local choiceList = {}
		 for c = 1, numChoices, 1 do
				local name, texture, numItems, quality, isUsable = GetQuestLogChoiceInfo(c)
				local itemLink = GetQuestLogItemLink("choice", c)
				table.insert(choiceList, {
				 name = name,
				 texture = texture,
				 numItems = numItems,
				 quality = quality,
				 isUsable = isUsable,
				 itemLink = itemLink,
				})
		 end
		 
		 local spellTexture, spellName, spellIsTradeskillSpell, spellIsSpellLearned = GetQuestLogRewardSpell()
		 
		 local numRewards = GetNumQuestLogRewards()
		 local rewardList = {}
		 for c = 1, numRewards, 1 do
			 local name, texture, numItems, quality, isUsable = GetQuestLogRewardInfo(c)
			 local itemLink = GetQuestLogItemLink("reward", c)
			 table.insert(rewardList, {
			  name = name,
			  texture = texture,
			  numItems = numItems,
			  quality = quality,
			  isUsable = isUsable,
				 itemLink = itemLink,
			 })
		 end

   table.insert(o.quests, {
    questId = tonumber(GetQuestLink(i):match(":(%d+):")),
    zone = zone,
    questTitle = questTitle,
    level = level,
    questTag = questTag,
    suggestedGroup = suggestedGroup,
    isComplete = isComplete,
    savedIsComplete = isComplete,
    isDaily = isDaily,
    questDescription = questDescription,
    questObjectives = questObjectives,
    numObjectives = numObjectives,
    objectiveList = objectiveList,
    money = GetQuestLogRequiredMoney(),
    numRewards = numRewards,
	   numChoices = numChoices,
	   moneyReward = GetQuestLogRewardMoney(),
	   playerTitle = GetQuestLogRewardTitle(),
	   isFailed = IsCurrentQuestFailed(),
	   myMoney = GetMoney(),
	   material = GetQuestBackgroundMaterial(),
	   choiceList = choiceList,
	   spellTexture = spellTexture,
	   spellName = spellName,
	   spellIsTradeskillSpell = spellIsTradeskillSpell,
	   spellIsSpellLearned = spellIsSpellLearned,
	   rewardList = rewardList,
	   questLink = GetQuestLink(i),
	   pushable = GetQuestLogPushable(),
	   talents = GetQuestLogRewardTalents(),
	   isWatched = IsQuestWatched(i),
   })
  else
   zone = questTitle or ""
  end
 end
 SelectQuestLogEntry(originalEntry)
 myquests = o
 return o
end
